package anywheresoftware.b4a.agraham.byteconverter;

import java.io.UnsupportedEncodingException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.CharBuffer;
import java.nio.DoubleBuffer;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.nio.LongBuffer;
import java.nio.ShortBuffer;
import java.nio.charset.Charset;
import java.util.Set;
import java.util.SortedMap;

import anywheresoftware.b4a.BA.Author;
import anywheresoftware.b4a.BA.ShortName;
import anywheresoftware.b4a.BA.Version;

/**
 *This ByeConverter object can transform arrays of primitive types to and from arrays
 *of Bytes with the "endian-ness" of the conversion specified. 
 *
 *Byte arrays can be transformed to and from a hexadecimal string representation.
 *
 *Strings can be transformed to and from arrays of Bytes using specified encodings and
 *to and from arrays of Chars.
 *
 *An array copy method is thrown in for good measure.
 */
@ShortName("ByteConverter")
@Author("Andrew Graham")
@Version(1.1f)
public class ByteConverter
{
	// 1.0 initial release
	// 1.1 HexToBytes now accepts lower case letters
	private static final double version = 1.1;

	/**
	 *This library contains a ByeConverter object that can transform arrays of primitive
	 *types to and from arrays of Bytes specifying the "endian-ness" of the conversion
	 *and also provides for the copying of part or all of the contents of one array to
	 *another array of the same type. 
	 *
	 *Byte arrays can be transformed to and from a hexadecimal string representation and
	 *Strings can be transformed to and from arrays of Bytes using specified encodings.
	 */
	public void LIBRARY_DOC()
	{
	}

	private char[] hexArray = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F' };

	private ByteOrder endian = ByteOrder.BIG_ENDIAN;

	// Public methods

	/**
	 *Gets or sets the endian-ness of byte conversions to and from other primitive values.
	 *The default for Android is Big_Endian, Windows PCs are Little_Endian.
	 */
	public boolean getLittleEndian()
	{
		return endian == ByteOrder.LITTLE_ENDIAN;
	}

	public void setLittleEndian(boolean littleEndian)
	{
		if (littleEndian)
			endian = ByteOrder.LITTLE_ENDIAN;
		else
			endian = ByteOrder.BIG_ENDIAN;
	}

	// char

	/**
	 *Takes an array of Bytes and returns an array of Chars converted from that array.
	 */
	public char[] CharsFromBytes(byte[] bytes)
	{
		CharBuffer cb = ByteBuffer.wrap(bytes).order(endian).asCharBuffer();
		char[] ca = new char[bytes.length / 2];
		cb.get(ca);
		return ca;

	}

	/**
	 * Takes an array of Chars and returns an array of Bytes converted from that array.
	 */
	public byte[] CharsToBytes(char[] vals)
	{
		ByteBuffer bb = ByteBuffer.allocate(vals.length * 2).order(endian);
		CharBuffer cb = bb.asCharBuffer();
		cb.put(vals);
		return bb.array();
	}

	// float

	/**
	 *Takes an array of Bytes and returns an array of Floats converted from that  array.
	 */
	public float[] FloatsFromBytes(byte[] bytes)
	{
		FloatBuffer fb = ByteBuffer.wrap(bytes).order(endian).asFloatBuffer();
		float[] fa = new float[bytes.length / 4];
		fb.get(fa);
		return fa;
	}

	/**
	 *Takes an array of Floats and returns an array of Bytes converted from that array.
	 */
	public byte[] FloatsToBytes(float[] vals)
	{
		ByteBuffer bb = ByteBuffer.allocate(vals.length * 4).order(endian);
		FloatBuffer fb = bb.asFloatBuffer();
		fb.put(vals);
		return bb.array();
	}

	// double

	/**
	 *Takes an array of Bytes and returns an array of Doubles converted from that array.
	 */
	public double[] DoublesFromBytes(byte[] bytes)
	{
		DoubleBuffer db = ByteBuffer.wrap(bytes).order(endian).asDoubleBuffer();
		double[] da = new double[bytes.length / 8];
		db.get(da);
		return da;
	}

	/**
	 *Takes an array of Doubles and returns an array of Bytes converted from that array.
	 */
	public byte[] DoublesToBytes(double[] vals)
	{
		ByteBuffer bb = ByteBuffer.allocate(vals.length * 8).order(endian);
		DoubleBuffer db = bb.asDoubleBuffer();
		db.put(vals);
		return bb.array();
	}

	// short

	/**
	 *Takes an array of Bytes and returns an array of Shorts converted from that array.
	 */
	public short[] ShortsFromBytes(byte[] bytes)
	{
		ShortBuffer sb = ByteBuffer.wrap(bytes).order(endian).asShortBuffer();
		short[] sa = new short[bytes.length / 2];
		sb.get(sa);
		return sa;
	}

	/**
	 *Takes an array of Shorts and returns an array of Bytes converted from that array.
	 */
	public byte[] ShortsToBytes(short[] vals)
	{
		ByteBuffer bb = ByteBuffer.allocate(vals.length * 2).order(endian);
		ShortBuffer sb = bb.asShortBuffer();
		sb.put(vals);
		return bb.array();
	}

	// int

	/**
	 *Takes an array of Bytes and returns an array of Ints converted from that array.
	 */
	public int[] IntsFromBytes(byte[] bytes)
	{
		IntBuffer ib = ByteBuffer.wrap(bytes).order(endian).asIntBuffer();
		int[] ia = new int[bytes.length / 4];
		ib.get(ia);
		return ia;
	}

	/**
	 *Takes an array of Ints and returns an array of Bytes converted from that array.
	 */
	public byte[] IntsToBytes(int[] vals)
	{
		ByteBuffer bb = ByteBuffer.allocate(vals.length * 4).order(endian);
		IntBuffer ib = bb.asIntBuffer();
		ib.put(vals);
		return bb.array();
	}

	// long

	/**
	 *Takes an array of Bytes and returns an array of Longs converted from that array.
	 */
	public long[] LongsFromBytes(byte[] bytes)
	{
		LongBuffer lb = ByteBuffer.wrap(bytes).order(endian).asLongBuffer();
		long[] la = new long[bytes.length / 8];
		lb.get(la);
		return la;
	}

	/**
	 *Takes an array of Longs and returns an array of Bytes converted from that array.
	 */
	public byte[] LongsToBytes(long[] vals)
	{
		ByteBuffer bb = ByteBuffer.allocate(vals.length * 8).order(endian);
		LongBuffer lb = bb.asLongBuffer();
		lb.put(vals);
		return bb.array();
	}

	// ArrayCopy

	/**
	 *Copies count elements from the srcoffset in the src array to destoffset in the dest array.
	 *The arrays need to be of the same type otherwise an ArrayStoreException is thrown.
	 */
	public void ArrayCopy(Object src, int srcOffset, Object dest, int destOffset, int count)
	{
		System.arraycopy(src, srcOffset, dest, destOffset, count);
	}

	// Hex string

	/**
	 *Returns a string containing the representation in hexadecimal of the bytes in a byte array.
	 */
	public String HexFromBytes(byte[] bytes)
	{
		char[] hexChars = new char[bytes.length * 2];
		int j = 0;
		for (int i = 0; i < bytes.length; i++)
		{
			int v = bytes[i] & 0xFF;
			hexChars[j++] = hexArray[v >> 4];
			hexChars[j++] = hexArray[v & 0x0f];
		}
		return new String(hexChars);
	}

	/**
	 *Returns an array of bytes containing the data represented by a hexadecimal string. The string can be upper or lower case.
	 */
	public byte[] HexToBytes(String hex) throws Exception
	{
		if (hex.length() % 2 > 0)
			throw new Exception("hex string has odd number of characters");
		byte[] bytes = new byte[hex.length() / 2];
		int i = 0, count = 0;
		boolean next = true;
		byte val = 0;
		try
		{
			for (i = 0; i < hex.length(); i++)
			{
				int chr = hex.charAt(i) - '0';
				if (chr > 9)
					chr -= 7;
				if (chr > 15)
					chr -= 32;
				val = (byte) ((val << 4) + chr);
				next = !next;
				if (next)
				{
					bytes[count] = val;
					count++;
				}
			}
			return bytes;
		} catch (RuntimeException e)
		{
			throw new Exception("Invalid character '" + hex.charAt(i) + "' in hex string");
		}
	}

	// Code page conversion

	/**
	 *Returns a string array containing all the values for encodings that are supported by this system.
	 */
	public String[] SupportedEncodings()
	{
		SortedMap<String, Charset> codings = Charset.availableCharsets();
		Set<?> set = codings.keySet();
		String[] sa = set.toArray(new String[0]);
		return sa;
	}

	/**
	 *Returns a string containing the bytes in the array interpreted as characters according to encoding. 
	 */
	public String StringFromBytes(byte[] bytes, String encoding)
	{
		try
		{
			return new String(bytes, 0, bytes.length, encoding);
		} catch (UnsupportedEncodingException e)
		{
			throw new RuntimeException(e);
		}
	}

	/**
	 *Returns an array of bytes containing the the characters in str converted to bytes according to encoding.
	 */
	public byte[] StringToBytes(String str, String encoding)
	{
		try
		{
			return str.getBytes(encoding);
		} catch (UnsupportedEncodingException e)
		{
			return new byte[0];
		}
	}

	/**
	 * Returns a string converted from the specified array of characters.
	 */
	public String FromChars(char[] chars)
	{
		return new String(chars);
	}

	/**
	 *Returns an array of chars comprising the characters of the specified string
	 */
	public char[] ToChars(String str)
	{
		return str.toCharArray();
	}

	// Public properties

	/**
	 *Returns the version number of this library.
	 */
	public double getVersion()
	{
		return version;
	}

}
